#!/usr/bin/python3

#SunPraise - WrathOfAinz - XenoTheStrange

debug = True

import socket
import contextlib

def connect_to_server(hostname, port, timeout=5.0):
    global debug
    """Create a tcp socket to a remote port. Return the socket if sucessful."""
    if not hostname:
        raise ValueError("Hostname cannot be empty")
    if not (0 <= port <= 65535):
        raise ValueError("Port must be between 0 and 65535")

    # Create socket with context manager for proper cleanup
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(timeout)
        try:
            # Connect to the server
            sock.connect((hostname, port))
            # Create a new socket to return (since context manager will close the original)
            connected_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            connected_sock.settimeout(timeout)
            connected_sock.connect((hostname, port))
            if debug:
                print(f"[CONNECT] SUCCESS ({hostname}:{port})")
            return connected_sock
        except (socket.gaierror, socket.timeout, socket.error) as e:
            raise ConnectionError(f"[CONNECT] FAIL ({hostname}:{port})")

def check_if_port_is_open(hostname, port):
    """Connects to a server and immediately disconnects if that succeeds, returns True if connected, False if failed."""
    try:
        sock = connect_to_server(hostname, port)
        sock.close()
        print(f"[DETECT] Port is open at ({hostname}:{port})")
        return True
    except (ValueError, socket.gaierror, socket.timeout, ConnectionError) as e:
        print(e, file=sys.stderr)
        return False

def make_status_request_packet():
    """
    Create a Minecraft Status Request packet (Packet ID 0x00 in Status state).

    Returns:
        bytes: The complete Status Request packet, including length prefix
    """
    # Packet ID: 0x00 (VarInt), no additional fields
    packet = bytearray()
    packet.extend(write_varint(0x00))

    # Prefix with packet length (VarInt)
    final_packet = bytearray()
    final_packet.extend(write_varint(len(packet)))
    final_packet.extend(packet)

    return bytes(final_packet)

def server_list_ping(sock, protocol_version, server_address, server_port):
    """
    Perform a Minecraft Server List Ping to get server status.

    Args:
        sock (socket.socket): Connected socket (e.g., from connect_to_server)
        protocol_version (int): Protocol version (e.g., 770 for Minecraft 1.21.5)
        server_address (str): Hostname or IP (e.g., 'localhost')
        server_port (int): Server port (e.g., 25565)

    Returns:
        str: The JSON string from the Status Response, or None if failed

    Raises:
        ValueError: If inputs are invalid
        ConnectionError: If socket operations fail
    """
    try:
        # Step 1: Send Handshake packet (next_state=1 for Status)
        handshake_packet = make_handshake_packet(
            protocol_version, server_address, server_port, next_state=1
        )
        sock.send(handshake_packet)

        # Step 2: Send Status Request packet
        status_request_packet = make_status_request_packet()
        sock.send(status_request_packet)

        # Step 3: Receive Status Response
        # Read enough bytes to get the length VarInt (up to 5 bytes for max VarInt)
        buffer = b""
        while len(buffer) < 5:
            chunk = sock.recv(4096)
            if not chunk:
                raise ConnectionError("Server closed connection")
            buffer += chunk

        # Parse packet length
        packet_length, bytes_read = read_varint(buffer)
        if packet_length <= 0:
            raise ValueError("Invalid packet length")

        # Read the full packet
        remaining_bytes = packet_length
        packet_data = buffer[bytes_read:]
        while len(packet_data) < remaining_bytes:
            chunk = sock.recv(4096)
            if not chunk:
                raise ConnectionError("Server closed connection")
            packet_data += chunk

        # Parse packet: Packet ID (VarInt) + JSON string (length-prefixed)
        packet_id, id_bytes_read = read_varint(packet_data)
        if packet_id != 0x00:
            raise ValueError(f"Expected Status Response (ID 0x00), got ID {packet_id}")

        # Read JSON string length (VarInt)
        json_length, json_length_bytes = read_varint(packet_data[id_bytes_read:])
        json_data = packet_data[id_bytes_read + json_length_bytes:id_bytes_read + json_length_bytes + json_length]

        # Decode JSON string (UTF-8)
        return json_data.decode('utf-8')

    except (ValueError, ConnectionError) as e:
        print(f"[PING] FAIL ({server_address}:{server_port}): {e}")
        return None

def read_varint(byte_stream):
    """
    Read a VarInt from a byte stream.

    Args:
        byte_stream (bytes): The byte stream to read from (e.g., from a socket)

    Returns:
        tuple: (int, int) - The decoded VarInt value and number of bytes read

    Raises:
        ValueError: If the VarInt is too big or the stream is exhausted
    """
    value = 0
    position = 0
    bytes_read = 0

    while True:
        if bytes_read >= len(byte_stream):
            raise ValueError("Byte stream exhausted while reading VarInt")

        current_byte = byte_stream[bytes_read]
        bytes_read += 1
        value |= (current_byte & 0x7F) << position

        if (current_byte & 0x80) == 0:
            return value, bytes_read

        position += 7

        if position >= 32:
            raise ValueError("VarInt is too big")

def write_varint(value):
    """
    Encode an integer as a VarInt.

    Args:
        value (int): The integer to encode as a VarInt (allows -1 for protocol version probing)

    Returns:
        bytes: The VarInt-encoded bytes

    Raises:
        ValueError: If the value is negative (except -1) or too large for VarInt
    """
    if value < -1 or value > 0x7FFFFFFF:
        raise ValueError("VarInt must be -1 or non-negative and fit in a VarInt")

    # Handle -1 (encoded as 0xFFFFFFFF)
    if value == -1:
        return b'\xFF\xFF\xFF\xFF\xFF'

    result = bytearray()

    while True:
        if (value & ~0x7F) == 0:
            result.append(value)
            return bytes(result)

        result.append((value & 0x7F) | 0x80)
        value >>= 7

def make_handshake_packet(protocol_version, server_address, server_port, next_state):
    """
    Create a Minecraft serverbound handshake packet.

    Args:
        protocol_version (int): Protocol version number (e.g., 770 for Minecraft 1.21.5, -1 for probing)
        server_address (str): Hostname or IP (e.g., 'localhost' or '127.0.0.1'), max 255 chars
        server_port (int): Server port (default 25565), must be 0-65535
        next_state (int): Next state (1 for Status, 2 for Login, 3 for Transfer)

    Returns:
        bytes: The complete handshake packet, including length prefix

    Raises:
        ValueError: If inputs are invalid (e.g., invalid port, invalid next_state, etc.)
    """
    # Input validation
    if protocol_version < -1 or protocol_version > 0x7FFFFFFF:
        raise ValueError("Protocol version must be -1 or non-negative and fit in a VarInt")
    if not isinstance(server_address, str) or len(server_address) > 255:
        raise ValueError("Server address must be a string of 255 or fewer characters")
    if not (0 <= server_port <= 65535):
        raise ValueError("Server port must be between 0 and 65535")
    if next_state not in (1, 2, 3):
        raise ValueError("Next state must be 1 (Status), 2 (Login), or 3 (Transfer)")

    # Build packet data (without length prefix)
    packet = bytearray()

    # Packet ID: 0x00 (VarInt)
    packet.extend(write_varint(0x00))

    # Protocol Version (VarInt)
    packet.extend(write_varint(protocol_version))

    # Server Address (String: VarInt length + UTF-8 bytes)
    address_bytes = server_address.encode('utf-8')
    if len(address_bytes) > 255:
        raise ValueError("Server address UTF-8 encoding exceeds 255 bytes")
    packet.extend(write_varint(len(address_bytes)))
    packet.extend(address_bytes)

    # Server Port (Unsigned Short, 2 bytes, big-endian)
    packet.extend(server_port.to_bytes(2, byteorder='big'))

    # Next State (VarInt Enum)
    packet.extend(write_varint(next_state))

    # Prefix with packet length (VarInt)
    final_packet = bytearray()
    final_packet.extend(write_varint(len(packet)))
    final_packet.extend(packet)

    return bytes(final_packet)


import ipaddress
from concurrent.futures import ThreadPoolExecutor
import threading

# Global lock for thread-safe file writing

def generate_ip_range(ip_address, num_octets=2):
    """
    Generate all IP addresses in the range where the first num_octets match the input IP.

    Args:
        ip_address (str): The input IP address (e.g., '104.36.87.135')
        num_octets (int, optional): Number of octets to fix (1-3, default: 2)

    Yields:
        str: Each IP address in the range (e.g., '104.36.0.0' to '104.36.255.255' for num_octets=2)

    Raises:
        ValueError: If the IP address is invalid or num_octets is invalid
    """
    try:
        # Parse the input IP address
        ip = ipaddress.ip_address(ip_address)
        if not isinstance(ip, ipaddress.IPv4Address):
            raise ValueError("Only IPv4 addresses are supported")

        # Validate num_octets
        if not (1 <= num_octets <= 3):
            raise ValueError("num_octets must be between 1 and 3")

        # Extract the fixed octets
        octets = ip_address.split('.')
        if len(octets) != 4:
            raise ValueError("Invalid IPv4 address format")
        base = '.'.join(octets[:num_octets])

        # Generate IPs based on num_octets
        if num_octets == 1:
            for i in range(256):
                for j in range(256):
                    for k in range(256):
                        yield f"{base}.{i}.{j}.{k}"
        elif num_octets == 2:
            for i in range(256):
                for j in range(256):
                    yield f"{base}.{i}.{j}"
        elif num_octets == 3:
            for i in range(256):
                yield f"{base}.{i}"

    except ValueError as e:
        raise ValueError(f"Invalid IP address or parameters: {e}")

def scan_ip_range(ip_address, port=25565, max_threads=100, num_octets=2, timeout=2.0):
    """
    Scan a range of IP addresses in parallel, calling do_get_server_status on each.

    Args:
        ip_address (str): The input IP address to base the range on (e.g., '104.36.87.135')
        port (int, optional): The port to scan (default: 25565)
        max_threads (int, optional): Maximum number of concurrent threads (default: 100)
        num_octets (int, optional): Number of octets to fix in the IP range (1-3, default: 2)
        timeout (float, optional): Timeout in seconds for each connection attempt (default: 2.0)

    Returns:
        list: List of IP addresses where do_get_server_status returned True

    Raises:
        ValueError: If the IP address, port, num_octets, or timeout is invalid
    """
    if not (0 <= port <= 65535):
        raise ValueError("Port must be between 0 and 65535")
    if max_threads < 1:
        raise ValueError("max_threads must be at least 1")
    if not (1 <= num_octets <= 3):
        raise ValueError("num_octets must be between 1 and 3")
    if timeout <= 0:
        raise ValueError("timeout must be positive")

    # List to store successful IPs
    successful_ips = []

    def check_ip(ip):
        """
        Helper function to check a single IP with thread-safe file writing.
        """
        # Assume do_get_server_status is defined elsewhere
        try:
            # Wrap file-writing in a lock to prevent concurrent access
            result = do_get_server_status(ip, port, timeout=timeout)
            if result:
                return ip
        except Exception as e:
            # Log errors but don't fail the entire scan
            print(f"[SCAN] FAIL ({ip}:{port}): {e}")
            return None

    # Generate IPs and scan in parallel
    with ThreadPoolExecutor(max_workers=max_threads) as executor:
        # Map check_ip to all IPs in the range
        results = executor.map(check_ip, generate_ip_range(ip_address, num_octets))

        # Collect successful IPs
        for result in results:
            if result:
                successful_ips.append(result)

    return successful_ips

def append_to_file(file, data):
    with open(file, "a") as file:
        file.write(data)
    return

def write_file(file, data):
    with open(file, "w") as file:
        file.write(data)
    return

def do_get_server_status(hostname, port, timeout=3):
    try:
        sock = connect_to_server(hostname, port, timeout=timeout)
        # Perform Server List Ping
        json_response = server_list_ping(
            sock, protocol_version=770, server_address=hostname, server_port=port
        )
        if json_response:
            print("Server Status:", json_response)
            #append_to_file("./alive_servers.txt", f"{hostname}:{port}\n")
            write_file(f"./logs/{hostname}.{port}.txt", json_response)
        else:
            print("Failed to get server status")
        sock.close()
        return True
    except ConnectionError as e:
        print(f"Failed to connect to ({hostname}:{port})")
        #print("CONNECTION FAILED")
        return False

import argparse
import sys
import functools
print = functools.partial(print, flush=True)

def parse_args():
    parser = argparse.ArgumentParser(description="Connect to a remote server via socket")
    parser.add_argument("hostname", type=str, help="Hostname or IP address of the remote server")
    parser.add_argument("-p", "--port", type=int, default=25565, help="Port number (default: 25565)")
    parser.add_argument("-o3", "--octet-3",action="store_true",default=False,help="Also scan the third octet IPs, ex 54.*.*.* without overloading memory.")
    parser.add_argument("-t", "--threads", type=int, default=256, help="Number of worker threads for multithreaded scans. Set to '1' to disable multithreading.")
    return parser.parse_args()

def main(args):
    #check_if_port_is_open(args.hostname, args.port)
    #do_get_server_status(args.hostname, args.port)
    if args.octet_3 == True:
        sname = args.hostname.split(".")
        start = int(sname[1])
        ipTemplate = f"{sname[0]}.THISIP.{sname[2]}.{sname[3]}"
        for newIP in range(start, 256):
            thisIP = ipTemplate.replace("THISIP", str(newIP))
            successful_ips = scan_ip_range(thisIP, port=args.port, max_threads=args.threads, num_octets=2)
            print("Servers found at:", successful_ips)
    else:
        successful_ips = scan_ip_range(args.hostname, port=args.port, max_threads=args.threads, num_octets=2) #setting octets to 3 causes memory-related crashing.
        print("Servers found at:", successful_ips)
    print("Scan finished")

if __name__ == "__main__":
    args = parse_args()
    main(args)
